#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "nrf.h"
#include "app_error.h"

#include "nrf_log.h"

#include "fds.h"
#include "nrf_atfifo.h"

#include "app_fds.h"

/*
***************************************************************************
                    FDS RELATED PREPROCESOR DIRECTIVES
***************************************************************************
*/

/* BOOT INIT File Id and Record Key */
#define BOOTINIT_FILE     (0x8010)
#define BOOTINIT_REC_KEY  (0x7010)


/* CONFIG File Id and Record Key */
#define CONFIG_FILE     (0x8011)
#define CONFIG_REC_KEY  (0x7011)

/* SAADC File Id and Record Key */
#define SAADC_FILE     (0x8012)
#define SAADC_REC_KEY  (0x7012)

// Values of the application configuration
#define APP_ID              1
#define APP_VERSION         (0x01A0)        // Version 1.0
#define APP_RECORD_VERSION  (0x01A0)        // Version 1.0
#define APP_NODE_NUMBER     0
#define APP_NODE_NAME       "Nameless"

/*
***************************************************************************
                    FDS RELATED GLOBAL VARIABLES
***************************************************************************
*/

/* *********************** BOOT INIT FILE ******************************* */
// Constant varibale containig the boot init file
static app_bootinit_data_t m_fds_bootinit_data =
{
    .boot_count = 0,
    .last_error_code = 0,
};

// Constant varibale containig the record
static fds_record_t const m_fds_bootinit_record =
{
    .file_id           = BOOTINIT_FILE,
    .key               = BOOTINIT_REC_KEY,
    .data.p_data       = &m_fds_bootinit_data,
    /* The length of a record is always expressed in 4-byte units (words). */
    .data.length_words = (sizeof(m_fds_bootinit_data) + 3) / sizeof(uint32_t),
};

/* *********************** APP CONFIG FILE ****************************** */
// Constant varibale containig the aplication configuration
static app_configuration_data_t m_fds_app_cfg_data =
{
    .app_id = APP_ID,
    .app_version = APP_VERSION,
    .record_version = APP_RECORD_VERSION,
    .node_number = APP_NODE_NUMBER,
    .node_name = APP_NODE_NAME,
};

// Constant varibale containig the record
static fds_record_t const m_fds_app_cfg_record =
{
    .file_id           = CONFIG_FILE,
    .key               = CONFIG_REC_KEY,
    .data.p_data       = &m_fds_app_cfg_data,
    /* The length of a record is always expressed in 4-byte units (words). */
    .data.length_words = (sizeof(m_fds_app_cfg_data) + 3) / sizeof(uint32_t),
};


// Temp data to store config data that can be modified by user
static uint32_t tmp_node_number;
static char     tmp_node_name[sizeof(m_fds_app_cfg_data.node_name)] = {0};
static bool     tmp_node_number_change = false;
static bool     tmp_node_name_change = false;

/* Global variable for detecting FDS event and its result */
volatile static bool m_fds_event;
volatile static ret_code_t m_fds_event_code;
volatile static fds_evt_id_t m_fds_event_id;

/*
*********************************************************************
                            FDS LIBRARY
*********************************************************************
*/

static void fds_evt_handler(fds_evt_t const * p_evt)
{
    m_fds_event = true;
    m_fds_event_code = p_evt->result;
    m_fds_event_id = p_evt->id;

    if (p_evt->result == NRF_SUCCESS)
    {
        NRF_LOG_INFO("[FDS] Event: %i received (NRF_SUCCESS)",
                      p_evt->id);
    }
    else
    {
        NRF_LOG_INFO("[FDS] Event: %i received code error (%u)",
                      p_evt->id,
                      p_evt->result);
    }

    switch (p_evt->id)
    {
        case FDS_EVT_INIT:
            if (p_evt->result != NRF_SUCCESS)
            {
                NRF_LOG_INFO("[FDS](I) Init ERROR (code: %u)", p_evt->result);
            }
            break;
        case FDS_EVT_WRITE:
            if (p_evt->result == NRF_SUCCESS)
            {
                NRF_LOG_INFO("[FDS](W) Record ID:\t0x%04x",  p_evt->write.record_id);
                NRF_LOG_INFO("[FDS](W) File ID:\t0x%04x",    p_evt->write.file_id);
                NRF_LOG_INFO("[FDS](W) Record key:\t0x%04x", p_evt->write.record_key);
                NRF_LOG_INFO("[FDS](W) Is Record Updated: \t%c", (p_evt->write.is_record_updated == true) ? 'Y' : 'N');
            }
            else
            {
                NRF_LOG_INFO("[FDS](W) Write ERROR (code: %u)", p_evt->result);
            }
            break;
        case FDS_EVT_UPDATE:
            if (p_evt->result == NRF_SUCCESS)
            {
                NRF_LOG_INFO("[FDS](U) Record ID:\t0x%04x",  p_evt->write.record_id);
                NRF_LOG_INFO("[FDS](U) File ID:\t0x%04x",    p_evt->write.file_id);
                NRF_LOG_INFO("[FDS](U) Record key:\t0x%04x", p_evt->write.record_key);
                NRF_LOG_INFO("[FDS](U) Is Record Updated:\t%c", (p_evt->write.is_record_updated == true) ? 'Y' : 'N');
            }
            else
            {
                NRF_LOG_INFO("[FDS](U) Update ERROR (code: %u)", p_evt->result);
            }
            break;
        case FDS_EVT_GC:
            // Garbage collection done
            if (p_evt->result == NRF_SUCCESS)
            {
                NRF_LOG_INFO("[FDS](G) Garbage Collection done (NRF_SUCCESS)");
            }
            else
            {
                NRF_LOG_INFO("[FDS](G) Garbage Collection ERROR (code: %u)", p_evt->result);
            }
            break;
        /*
        case FDS_EVT_DEL_RECORD:
        {
            if (p_evt->result == NRF_SUCCESS)
            {
                NRF_LOG_INFO("Record ID:\t0x%04x",  p_evt->del.record_id);
                NRF_LOG_INFO("File ID:\t0x%04x",    p_evt->del.file_id);
                NRF_LOG_INFO("Record key:\t0x%04x", p_evt->del.record_key);
            }
            m_delete_all.pending = false;
        } break;
        */

        default:
            break;
    }
}


/*
* Brief: Load boot-init file.
*        - If file does not exist then exit with error
*        - If file exist then load it, and put data into *pconfig structure
*        Return: error code from FDS library (NRF_SUCCESS if it was a success)
*/
ret_code_t app_fds_bootinit_load(app_bootinit_data_t *pconfig)
{
    ret_code_t rc;
    fds_record_desc_t desc = {0};
    fds_find_token_t  tok  = {0};

    rc = fds_record_find(BOOTINIT_FILE, BOOTINIT_REC_KEY, &desc, &tok);

    if (rc == NRF_SUCCESS)
    {
        
        fds_flash_record_t frec = {0};

        rc = fds_record_open(&desc, &frec);
        if (rc == NRF_SUCCESS)
        {
            /* Copy the configuration from flash into pconfig. */
            memcpy(pconfig, frec.p_data, sizeof(app_bootinit_data_t));
            rc = fds_record_close(&desc);
        }
    }

    return rc;
}


/*
* Brief: Load configuration file.
*        - If file does not exist then exit with error
*        - If file exist then load it, and put data into *pconfig structure
*        Return: error code from FDS library (NRF_SUCCESS if it was a success)
*/
ret_code_t app_fds_config_load(app_configuration_data_t *pconfig)
{
    ret_code_t rc;
    fds_record_desc_t desc = {0};
    fds_find_token_t  tok  = {0};

    rc = fds_record_find(CONFIG_FILE, CONFIG_REC_KEY, &desc, &tok);

    if (rc == NRF_SUCCESS)
    {
        
        fds_flash_record_t frec = {0};

        rc = fds_record_open(&desc, &frec);
        if (rc == NRF_SUCCESS)
        {
            /* Copy the configuration from flash into pconfig. */
            memcpy(pconfig, frec.p_data, sizeof(app_configuration_data_t));
            rc = fds_record_close(&desc);
        }
    }

    return rc;
}

/*
* Brief: Update configuration file.
*        - If file does not exist then return with the error code
*        - If file exist then it perform an update
*        Return: error code from FDS library (NRF_SUCCESS if it was a success)
*/
ret_code_t app_fds_config_update(void)
{
    ret_code_t rc;
    fds_record_desc_t desc = {0};
    fds_find_token_t  tok  = {0};

    rc = fds_record_find(CONFIG_FILE, CONFIG_REC_KEY, &desc, &tok);

    if (rc == NRF_SUCCESS)
    {
        /* Write the updated record to flash. */
        rc = fds_record_update(&desc, &m_fds_app_cfg_record);
        if(rc == NRF_SUCCESS)
        {
            while(!m_fds_event);
            m_fds_event = false;
            if(m_fds_event_code == NRF_SUCCESS)
            {
                NRF_LOG_INFO("AU Record ID:\t0x%04x", desc.record_id);
                NRF_LOG_INFO("AU GC Count:\t%u", desc.gc_run_count);
                NRF_LOG_INFO("AU Record Open:\t%c", desc.record_is_open ? 'Y' : 'N');
                return NRF_SUCCESS;
            }
            else
            {
                APP_ERROR_CHECK(m_fds_event_code);
            }
            
        }
    }
    return rc;
}



/*
* Brief: Load, create or update boot-init file.
*        - If file does not exist then create new one with values in zero
*        - If file exist then load it, increase the boot count 
*          and update it (the porpuse of boot count is to track the number of times
*          the device was restared)
*        - If there is no space avaliable the GC (garbage collector) is triggered.
*       Return: error code defined by me.
*/
app_fds_res_t app_fds_bootinit_start(void)
{
    ret_code_t rc;
    fds_stat_t stat = {0};

    rc = fds_stat(&stat);
    APP_ERROR_CHECK(rc);

    NRF_LOG_INFO("Found %d valid records.", stat.valid_records);
    NRF_LOG_INFO("Found %d dirty records (ready to be garbage collected).", stat.dirty_records);

    fds_record_desc_t desc = {0};
    fds_find_token_t  tok  = {0};

    rc = fds_record_find(BOOTINIT_FILE, BOOTINIT_REC_KEY, &desc, &tok);

    if (rc == NRF_SUCCESS)
    {
        /* A config file is in flash. Let's update it. */
        fds_flash_record_t bootinit = {0};

        /* Open the record and read its contents. */
        rc = fds_record_open(&desc, &bootinit);
        APP_ERROR_CHECK(rc);

        /* Copy the configuration from flash into m_app_cfg. */
        memcpy(&m_fds_bootinit_data, bootinit.p_data, sizeof(app_bootinit_data_t));

        NRF_LOG_INFO("Boot-Init file found, updating boot count to %d.", m_fds_bootinit_data.boot_count);

        /* Update boot count. */
        m_fds_bootinit_data.boot_count++;

        /* Close the record when done reading. */
        rc = fds_record_close(&desc);
        APP_ERROR_CHECK(rc);

        /* Write the updated record to flash. */
        rc = fds_record_update(&desc, &m_fds_bootinit_record);
        if ((rc != NRF_SUCCESS) && (rc == FDS_ERR_NO_SPACE_IN_FLASH))
        {
            NRF_LOG_INFO("No space in flash, starting garbage collector...");
            rc = fds_gc();
            if(rc == NRF_SUCCESS)   // TODO: Verify if event ID equals to GC
            {
                // Wait for fds library response
                while(!m_fds_event);
                m_fds_event = false;
                if(m_fds_event_code == NRF_SUCCESS)
                {
                    return APP_FDS_GC_IN_PROGRESS;
                }
                else
                {
                    APP_ERROR_CHECK(m_fds_event_code);
                }
            }
            else APP_ERROR_CHECK(rc);
        }
        else
        {
            if(rc == NRF_SUCCESS)      // TODO: Verify if event ID equals to UPDATE
            {
                while(!m_fds_event);
                m_fds_event = false;
                if(m_fds_event_code == NRF_SUCCESS)
                {
                    NRF_LOG_INFO("BOOT Record ID:\t0x%04x", desc.record_id);
                    NRF_LOG_INFO("BOOT GC Count:\t%u", desc.gc_run_count);
                    NRF_LOG_INFO("BOOT Record open:\t%c", desc.record_is_open ? 'Y' : 'N');
                    return APP_FDS_OK;
                }
                else
                {
                    APP_ERROR_CHECK(m_fds_event_code);
                }
            }
            else APP_ERROR_CHECK(rc);
        }
    }
    else
    {
        /* System config not found; write a new one. */
        NRF_LOG_INFO("Boot-Init file not found, writing default file...");

        rc = fds_record_write(&desc, &m_fds_bootinit_record);
        if ((rc != NRF_SUCCESS) && (rc == FDS_ERR_NO_SPACE_IN_FLASH))
        {
            NRF_LOG_INFO("No space in flash, starting garbage collector...");
            rc = fds_gc();
            if(rc == NRF_SUCCESS)   // TODO: Verify if event ID equals to GC
            {
                // Wait for fds library response
                while(!m_fds_event);
                m_fds_event = false;
                if(m_fds_event_code == NRF_SUCCESS)
                {
                    return APP_FDS_GC_IN_PROGRESS;
                }
                else
                {
                    APP_ERROR_CHECK(m_fds_event_code);
                }
            }
            else APP_ERROR_CHECK(rc);
        }
        else
        {
            if(rc == NRF_SUCCESS)
            {
                // Wait for fds library response
                while(!m_fds_event);
                m_fds_event = false;
                if(m_fds_event_code == NRF_SUCCESS)
                {
                    NRF_LOG_INFO("BOOT Record ID:\t0x%04x", desc.record_id);
                    NRF_LOG_INFO("BOOT GC Count:\t%u", desc.gc_run_count);
                    NRF_LOG_INFO("BOOT Record key:\t%c", desc.record_is_open ? 'Y' : 'N');
                    return APP_FDS_OK;
                }
                else
                {
                    APP_ERROR_CHECK(m_fds_event_code);
                }
            }
            else APP_ERROR_CHECK(rc);
        }
    }

    return APP_FDS_FATAL_ERROR;
}


/*
* Brief: Load, create or update configuration file.
*        - If file does not exist then create new one wiht initial values
*        - If file exist then load it, increase the boot count 
*          and update it (the porpuse of boot count is to track the number of times
*          the device was restared)
*        - If there is no space avaliable the GC (garbage collector) is triggered.
*       Return: error code defined by me.
*/
app_fds_res_t app_fds_config_start(void)
{
    ret_code_t rc;

    fds_record_desc_t desc = {0};
    fds_find_token_t  tok  = {0};

    rc = fds_record_find(CONFIG_FILE, CONFIG_REC_KEY, &desc, &tok);

    if (rc == NRF_SUCCESS)
    {
        /* A config file is in flash. */
        fds_flash_record_t config = {0};

        /* Open the record and read its contents. */
        rc = fds_record_open(&desc, &config);
        APP_ERROR_CHECK(rc);

        /* Copy the configuration from flash into m_app_cfg. */
        memcpy(&m_fds_app_cfg_data, config.p_data, sizeof(app_configuration_data_t));

        NRF_LOG_INFO("Config file found and loaded OK");

        /* Close the record when done reading. */
        rc = fds_record_close(&desc);
        APP_ERROR_CHECK(rc);

        return APP_FDS_OK;
    }
    else
    {
        /* System config not found; write a new one. */
        NRF_LOG_INFO("Config file not found, writing default file...");

        rc = fds_record_write(&desc, &m_fds_app_cfg_record);
        if ((rc != NRF_SUCCESS) && (rc == FDS_ERR_NO_SPACE_IN_FLASH))
        {
            NRF_LOG_INFO("No space in flash, starting garbage collector...");
            rc = fds_gc();
            if(rc == NRF_SUCCESS)   // TODO: Verify if event ID equals to GC
            {
                // Wait for fds library response
                while(!m_fds_event);
                m_fds_event = false;
                if(m_fds_event_code == NRF_SUCCESS)
                {
                    return APP_FDS_GC_IN_PROGRESS;
                }
                else
                {
                    APP_ERROR_CHECK(m_fds_event_code);
                }
            }
            else APP_ERROR_CHECK(rc);
        }
        else
        {
            if(rc == NRF_SUCCESS)
            {
                // Wait for fds library response
                while(!m_fds_event);
                m_fds_event = false;
                if(m_fds_event_code == NRF_SUCCESS)
                {
                    NRF_LOG_INFO("APP CFG Record ID:\t0x%04x", desc.record_id);
                    NRF_LOG_INFO("APP CFG GC Count:\t%u", desc.gc_run_count);
                    NRF_LOG_INFO("APP CFG Record key:\t%c", desc.record_is_open ? 'Y' : 'N');
                    return APP_FDS_OK;
                }
                else
                {
                    APP_ERROR_CHECK(m_fds_event_code);
                }
            }
            else APP_ERROR_CHECK(rc);
        }
    }

    return APP_FDS_FATAL_ERROR;
}

/*
* Brief: Register the callback function and Init the FDS library
*        It waits until the init event is generated.
*        Return: error code defined by me.
*/
app_fds_res_t app_fds_register_init(void)
{
    ret_code_t ret_code;

    m_fds_event = false;
    /* Register first to receive an event when initialization is complete. */
    (void) fds_register(fds_evt_handler);
    // Init fds
    ret_code = fds_init();
    APP_ERROR_CHECK(ret_code);
    // Wait for init event
    while(!m_fds_event);
    m_fds_event = false;
    if(m_fds_event_code == NRF_SUCCESS)
    {
        return APP_FDS_OK;
    }
    else
    {
        APP_ERROR_CHECK(m_fds_event_code);
    }

    return APP_FDS_FATAL_ERROR;

}


/*
* Brief: Save node number in a temp variable for later storage
*        using app_fds_commit_cfg_data_RAM function
*/
void app_fds_set_node_number_in_RAM(uint32_t node_number)
{
    tmp_node_number = node_number;
    tmp_node_number_change = true;
}

/*
* Brief: Save node name in a temp variable for later storage
*        using app_fds_commit_cfg_data_RAM function
*/
bool app_fds_set_node_name_in_RAM(char *node_name)
{
    bool result = false;

    if (strlen(node_name) < sizeof(m_fds_app_cfg_data.node_name))
    {
        strcpy(tmp_node_name, node_name);
        tmp_node_name_change = true;
        result = true;
    }
    return result;
}

/*
* Brief: Save temp data to the actual RAM structure
*/
void app_fds_commit_cfg_data_RAM(void)
{
    if(tmp_node_number_change)
    {
        m_fds_app_cfg_data.node_number = tmp_node_number;
        tmp_node_number_change = false;
    }
    if(tmp_node_name_change)
    {
        strcpy(m_fds_app_cfg_data.node_name, tmp_node_name);
        tmp_node_name_change = false;
    }
}

/*
* Brief: get the max character size of node name
*/
uint32_t app_fds_get_node_name_size(void)
{
    return((uint32_t)sizeof(m_fds_app_cfg_data.node_name) - 1);
}


